require 'base.scripts.vec'

GameAPI = {}

local windows = {
--full
	full = box2:new(0, 1);
--split horizontal
	top = box2:new(0, 0, 1, .5);
	bottom = box2:new(0, .5, 1, 1);
--split vertical
	left = box2:new(0, 0, .5, 1);
	right = box2:new(.5, 0, 1, 1);
--quadrants
	topLeft = box2:new(0, 0, .5, .5);
	topRight = box2:new(.5, 0, 1, .5);
	bottomLeft = box2:new(0, .5, .5, 1);
	bottomRight = box2:new(.5, .5, 1, 1);
}

GameAPI.shareASingleViewport = false

-- returns a list of viewports, mapped to each local client
-- if an active client gets a nil viewport then it's assumed to not have a viewport
--- this is useful for shared-screen games
function GameAPI:setupViewports(numLocals)
	if self.shareASingleViewport then
		return windows.full
	end
	
	if numLocals == 1 then
		return windows.full
	elseif numLocals == 2 then
		return windows.top, windows.bottom
	elseif numLocals == 3 then
		return windows.top, windows.bottomLeft, windows.bottomRight
	elseif numLocals == 4 then
		return windows.topLeft, windows.topRight, windows.bottomLeft, windows.bottomRight
	end
end

GameAPI.lastSpawnPoint = nil

--[[
forward-declaring, eh?
It is tempting to change this to a class reference, start it as 'nil', and leave it to each game to define

in other news, all this is used for is
(a) finding an associated spawnobject (getStartSpawnObj)
(b) the subsequent pull of spawnClient, which just calls its spawningClass:new with that spawnobject passed
if these two functions are overridden then GameAPI.clientSpawnClass becomes moot
--]]
GameAPI.clientSpawnClass = nil

--non-api 
function GameAPI:getStartSpawnObj()
	if not self.clientSpawnClass then
		print('GameAPI:getStartSpawnObj failed to be provided with a client spawn class')
		return
	end

	self.lastSpawnPoint = spawnObjFind(self.clientSpawnClass.classname, self.lastSpawnPoint)

	if not self.lastSpawnPoint then
		print('GameAPI:getStartSpawnObj failed to find a spawn object of type '..self.clientSpawnClass)
	end

	return self.lastSpawnPoint
end

-- non-api
function GameAPI:createClientEntityFromSpawnObject(so)
	if not self.clientSpawnClass then
		print('GameAPI:createClientEntityFromSpawnObject failed to be provided with a client spawn class')
		return
	end
	
	local clientEnt = self.clientSpawnClass:spawnClient(so)
	
	if not clientEnt then
		print('GameAPI:createClientEntityFromSpawnObject failed to create entity of type ' .. self.clientSpawnClass)
		return
	end

	return clientEnt
end

-- entptr_t GameAPI::spawnPlayer(int conn)
-- in our lua function its gonna return nil on fail, an ent table otherwise
function GameAPI:spawnPlayer(conn)

	local so = self:getStartSpawnObj()
	if not so then
		print('GameAPI:spawnPlayer failed to find a spawn object')
		return
	end
	
	local clientEnt = self:createClientEntityFromSpawnObject(so)
	if not clientEnt then
		print('GameAPI:spawnPlayer failed to create entity')
		return
	end

	-- if the level script of 'onClientSpawn' exists then run it
	-- consider merging this with the gameapi? allow the level to override it? meh, too commplicated
	if onClientSpawn then
		onClientSpawn(clientEnt)
	end
	
	return clientEnt
end

-- void GameAPI::clientSetInput(entptr_t e, ClientInput *in)
function GameAPI:clientSetInput(e, ...)
	if not e then
		print("GameAPI:clientSetInput(): couldn't find client entity")
		return
	end
	if not istable(e.input) then
		print('clientSetInput needs an input structure!')
		return
	end
	e.input:fill(...)
end

