require 'base.scripts.hud'
require 'base.scripts.text'
require 'base.scripts.util'	--exec()
require 'base.scripts.clientside'	--ClientSide

--[[
TODO ...
pure lua menus are a huge performance hit
this means we'll have to
(a) reintroduce the c++ menu bindings
(b) negotiate viewport mapping and input mapping ...
--]]

class('Menu')
Menu.sysSize = vec2:new(60,40)	-- doesn't scale well with children... TODO onresize etc? ...
Menu.renderDepth = -100		-- static
Menu.borderTex = spriteLoad('menuborder.png')
Menu.borderTexUVRatio = .25
Menu.borderTexWidth = 1

-- defaults:
Menu.fontSize = vec2:new(1.5)
Menu.fontColor = vec4:new(1)
Menu.backgroundColor = vec4:new(1,1,1,1)

function Menu:ctor(args)
	-- a general means for modaless menus ...
	-- 1) keep track of a uid
	-- 2) hold a serverside active menu array
	-- send a 'store' on create and a 'clear' on delete with the menu's uid
	-- ... but the server makes a client call and the client creates menus ...
	--	how does the server know how to associate its client call with what ids are created ...

	-- handle 'parent' and 'client' first
	if args.parent then
		self.parent = args.parent
		self.client = assert(self.parent.client)
		if not self.parent.child then self.parent.child = table.new() end
		self.parent.child:insert(self)
		
		if args.isPrompt or self.isPrompt then
			if not self.parent.childPrompts then
				self.parent.childPrompts = table.new()
			end
			self.parent.childPrompts:insertUnique(self)
		end
	else
		self.client = assert(args.client)
		
		-- subclass ClientSide to have a dialog-based one?

		self.client.updateCallbacks:insertUnique(ClientSide.updateMenus)
		
		-- current input context menus
		if not self.client.menuKeyPrompt then
			self.client.menuKeyPrompt = table.new()
		end
		
		-- root-level menus
		if not self.client.menus then
			self.client.menus = table.new()
		end
		self.client.menus:insert(self)
	end
	-- map over all else nxt
	for k,v in pairs(args) do
		if k ~= 'parent'
		and k ~= 'client'
		then
			self[k] = args[k]
		end
	end
	-- and instanciate copies of what we need to
	self.pos = vec2:new(assert(self.pos))
	self.size = vec2:new(assert(self.size))
	if self.fontSize then self.fontSize = vec2:new(self.fontSize) end
	if self.fontColor then self.fontColor = vec4:new(self.fontColor) end
	if self.backgroundColor then self.backgroundColor = vec4:new(self.backgroundColor) end
	
	if self.hasPrompts then
		self.client.menuKeyPrompt:insert(self)
		self.client.menuKeyPromptIndex = self.firstPromptIndex or 1
	end

	-- parallel'ing entities - in case i merge this with it
	self.active = true
end

function Menu:delete()
	-- explicitly delete all children
	if self.child then
		while #self.child > 0 do
			self.child[1]:delete()
		end
	end

	self.active = false
	-- remove from the menu input contexts?
	-- lock?
	self.client.menuKeyPrompt:removeObject(self)
	-- remove from parent tree
	if self.parent then
		self.parent.child:removeObject(self)
		if self.parent.childPrompts then
			self.parent.childPrompts:removeObject(self)
		end
	else
		self.client.menus:removeObject(self)
	end
	-- call custom on-delete function
	if self.onDelete then self:onDelete() end
end

-- if the menu provides an "onOk" or an "onCancel" but no "mouseEvent" then this will be called 
--	it's to merge keyboard and 
function Menu:mouseEventOkCancel(input)
	if self.onOk and input.key.attack.isPressed then
		self:onOk(input)
	end
	if self.onCancel and input.key.alt.isPressed then
		self:onCancel(input)
	end
	return true
end

--[[
this handles client-side stuff and rendering
however the input is still missing...
client sends it to server separate of the entity system
for it to get here the server would have to send it back to the client through the entity system...
...or add some new exposures to the client ...

... but even with client exposure we have the issue that
input-handling for ClientView is now performed ...
1) on the client side (for menus)
2) on the server side (for game logic -- delegate entity, etc)
... and how should the two communicate, so that dialog input doesn't interfere with game input?
--]]
function Menu:update(input, ofs, captured)
	local wb = box2:new(self.pos, self.pos + self.size)
	if ofs then wb:addTo(ofs) end
	wb:divTo(Menu.sysSize)
	
	-- store for later
	local viewbox = viewGetBox()
	self.worldBBox = box2:new(wb):map(viewbox)

	-- notice that the mouse position is still in world coordinates
	-- maybe, just maybe, someone might want the menu local coordinates?
	-- both actually look pretty helpful...
	
	local mouseInBounds = input.mouse.pos[1] >= wb.min[1]
		and input.mouse.pos[1] <= wb.max[1]
		and input.mouse.pos[2] >= wb.min[2]
		and input.mouse.pos[2] <= wb.max[2]

	if not captured.did and self.preMouseEvent and mouseInBounds then
		captured.did = self:preMouseEvent(input)
	end

	-- clear it here and let the clicked-on children set it for the parent later
	self.selectedChild = nil
	
	if self.child then
		if not ofs then ofs = vec2:new() end
		ofs = ofs + self.pos
		for i = #self.child,1,-1 do
			local m = self.child[i]
			if m then 
				if m:update(input, ofs, captured) then break end
			end
		end
	end

	if not captured.did and mouseInBounds then
		if (self.mouseEvent or self.onOk or self.onCancel) then
			if self.mouseEvent then
				captured.did = self:mouseEvent(input)
			else
				captured.did = self:mouseEventOkCancel(input)
			end
		end
		if self.parent then
			self.parent.selectedChild = self
		end
	end
end

function Menu:display(ofs)

	local wb = box2:new(self.pos, self.pos + self.size)
	if ofs then wb:addTo(ofs) end

	local viewbox = viewGetBox()
	wb:divTo(Menu.sysSize):map(viewbox)

	local fontsize = vec2:new(self.fontSize)
	fontsize:mulTo(viewbox:size()):divTo(Menu.sysSize)
	-- at the moment we're set up to draw either tex, a solid rect
	--  or borderTex, in 3x3 pieces sliced at [0, .25, .75, 1]
	if self.tex then
		drawRect{
			color=self.backgroundColor;
			box=wb;
			depth=self.renderDepth;
			tex=self.tex;
			blend=BLEND.DECAL;
			postFBO=true;
		}
	elseif self.borderTex then
		local divs = {0, self.borderTexUVRatio, 1-self.borderTexUVRatio, 1}
		local wbdivsu = {0, self.borderTexWidth/self.size[1], (self.size[1]-self.borderTexWidth)/self.size[1], 1}
		if self.size[1] < 2*self.borderTexWidth then
			wbdivsu[2] = .5
			wbdivsu[3] = .5
		end
		local wbdivsv = {0, self.borderTexWidth/self.size[2], (self.size[2]-self.borderTexWidth)/self.size[2], 1}
		if self.size[2] < 2*self.borderTexWidth then
			wbdivsv[2] = .5
			wbdivsv[3] = .5
		end
		for i=1,#divs-1 do
			for j=1,#divs-1 do
				local tcbox = box2:new{
					min={divs[i], divs[j]},
					max={divs[i+1], divs[j+1]}
				}
				local wbfrac = box2:new{
					min={wbdivsu[i], wbdivsv[j]},
					max={wbdivsu[i+1], wbdivsv[j+1]}
				}
				wbfrac:map(wb)
				drawRect{
					color=self.backgroundColor;
					box=wbfrac;
					tcbox=tcbox;
					depth=self.renderDepth;
					tex=self.borderTex;
					blend=BLEND.DECAL;
					postFBO=true;
				}
			end
		end
	end

	-- then draw text ontop
	if self.text then
		drawText{
			pos=wb.min;
			size=fontsize;
			depth=self.renderDepth;
			str=self.text;
			color=self.fontColor;
			width=wb.max[1] - wb.min[1];
			blend=BLEND.DECAL;
		}
	end

	if not ofs then ofs = vec2:new() end
	ofs = ofs + self.pos
	if self.child then
		for i = 1,#self.child do
			local m = self.child[i]
			m:display(ofs)
		end
	end
end

menuCursor = spriteLoad('base/sprites/menucursor.png')

function ClientSide:updateMenus()

	-- well this is kind of a hack...
	-- it's using a serverside command to get input on the clientside ...
	-- i'm starting to think we should expose these objects via userdata ptrs -- and via reflection ...
	-- and then just expose distinct client objects ..
	local input = ClientInput:new()
	if self.entID then
		input:fill(entClientGetInput(self.entID))
	else
		input:fill(clientGetMenuInput())
	end


	local captured = {}	--pointers in scripting languages...
	if self.menus then
		-- gather input in the opposite order you render
		-- TODO - only handle mouse input here
		-- save keyboard for the "self.menuKeyPrompt"
		-- or maybe we should have both mouse and keyboard use "menuKeyPrompt" ...
		-- but we'd stil need to worry about handling it in-order with the rendering...
		for i = #self.menus,1,-1 do
			local m = self.menus[i]
			if m then	-- in case it gets deleted...
				m:update(input, nil, captured)
			end
		end
		for i = 1,#self.menus do
			local m = self.menus[i]
			m:display()
		end
	end
	
	do -- if not captured.did then	-- all this condition does is hide the cursor when the mouse is over the box...
		-- [[lock
		if self.menuKeyPrompt and #self.menuKeyPrompt > 0 then
			local currentFocus = self.menuKeyPrompt[#self.menuKeyPrompt]
			if currentFocus.childPrompts and #currentFocus.childPrompts > 0 then
				--]]
				local index = self.menuKeyPromptIndex or 1
				index = ((index - 1) % #currentFocus.childPrompts) + 1
				
				local m = currentFocus.childPrompts[index]
				if m.worldBBox then	-- sometimes it takes a frame to refresh?
					self.menuKeyPromptIndex = index
					
					-- draw a marker by the current menu
					local viewBox = viewGetBox()
					local size = vec2:new(2 / Menu.sysSize[1], 2 / Menu.sysSize[2]) * viewBox:size()
					local pos = vec2:new{m.worldBBox.min[1] - size[1], m.worldBBox.min[2]-.1}
					
					drawRect{
						color={1,1,1,1};
						box={min=pos; max=pos + size;};
						depth=m.renderDepth-.1;
						tex=menuCursor;
						blend=BLEND.DECAL;
						postFBO=true;
					}
					
					-- rather than context sets
					-- how about have flagged children
					-- and just register the parent menu?
					
					-- this is where mouse clicks and keyboard 'ok's should be kept separate maybe?
					if input.key.menuOK.isPressed and m.onOk then
						m:onOk()		-- it better be there?
					elseif input.key.menuCancel.isPressed and m.onCancel then
						m:onCancel()
					-- hmm, menu control... gotta get fixed positions 
					elseif input.key.up.isPressed then
						self.menuKeyPromptIndex = self.menuKeyPromptIndex - 1
					elseif input.key.down.isPressed then
						self.menuKeyPromptIndex = self.menuKeyPromptIndex + 1
					elseif input.key.left.isPressed then
						self.menuKeyPromptIndex = self.menuKeyPromptIndex - 1
					elseif input.key.right.isPressed then
						self.menuKeyPromptIndex = self.menuKeyPromptIndex + 1
					end
				end
			end
		end
	end
end
