require 'base.scripts.gameapi'
require 'bomberman.scripts.item'
require 'bomberman.scripts.break'
require 'bomberman.scripts.ai'

-- maybe this should go in serverstart / main ?
function GameAPI:initPlayerSpawns()
	if self.playerSpawns then return end

	self.playerSpawns = table.new()

	local first = spawnObjFind('PlayerStart', nil)
	if first then
		local so = first
		repeat
			self.playerSpawns:insert{so=so}
			so = spawnObjFind('PlayerStart', so)
		until not so or so._spawnobj== first._spawnobj
	end

	-- might as well make this here too - it's a table subset of playerSpawns
	self.aiPlayers = table.new()
end


function GameAPI:spawnPlayer(connIndex,localIndex)
	self:initPlayerSpawns()

	-- now find the next player start not used ...
	for k,v in ipairs(self.playerSpawns) do
		-- if a v.ent happens to not be active (then chances are it'll get filled soon .. hmm ...)
		-- then that means the ent proly disconnected
		-- TODO - GameAPI:disconnectPlayer()
		if not v.ent or not v.ent:isActive() or not v.ent:isa(Player) then
			v.ent = PlayerStart:spawnClient(v.so)
			return v.ent
		end
	end
end

function GameAPI:calcAvoids()
	local avoid = table.new()
	for i,v in ipairs(self.playerSpawns) do
		if v.so.classname == 'PlayerStart' then
			avoid:insert(readVec2(v.so.pos))
		end
	end
	return avoid
end

--[[ for all tiles in the screen make random blocks
or - for all the tiles in the map -
if they are of 'bombableProbable' type or something or 'bombable and might not exist'
then flip a coin
if heads then make the block

i bet i don't have map size exposed ...
i certainly don't have mapSetType or mapSetKey exposed ...
--]]
function GameAPI:spawnBreaks(avoid)
	local avoidDist = 2	--how far around player starts not to put breakables
	local breaks = table.new()
	local t = vec2:new()
	for i=0,mapGetWidth()-1 do
		for j=0,mapGetHeight()-1 do
			if mapGetType(i,j) == MAPTYPE.Empty then
				t:set(i+.5,j+.5)
				local tooClose = false
				for k,p in ipairs(avoid) do
					if (p - t):lenSq() < avoidDist*avoidDist then	
						tooClose = true
						break
					end
				end
				if not tooClose then
					-- place a breakable (at random with random item) at t
					do -- if math.random() < .8 then
						breaks:insert(Break:new{pos=table.concat(t,' ')})
					end
				end
			end
		end
	end
	return breaks
end

--[[
now if I remember right there are a fixed amount of items per-stage ...
should we have the user specify them by spawning them off-screen somewhere?
then we can read in spawnobjects here and move them to underneath breakables ...
meh.
(that would make them level-specific)
maps to pairs: the class, and how many of them should show up in the game.
--]]
function GameAPI:addItems(breaks)
	local items = {
		{FlameItem, 5},
		{BombItem, 10},
		{IncineratorItem, 1},
		{RedBombItem, 1},
	}
	local done = false
	for i,v in ipairs(items) do 
		local classobj = v[1]
		local count = v[2]
		for j=1,count do
			local bc = #breaks
			if bc == 0 then
				print('we ran out of breakables to hide items under!\nthis should be a fun game')
				done = true
				break;
			end
			-- pick a random breakable
			local bi = math.random(bc)
			local br = breaks[bi]
			-- and put this under it
			br.itemType = classobj
			-- and remove that breakable from the breaks list
			breaks:remove(bi)
		end
		if done then break end	-- prematurely done
	end
end

-- this means no split screen
GameAPI.shareASingleViewport = true

-- run on server init
-- post load level?
-- todo - make a gameapi for serverstart
-- leaving 'main' out there for levels to use (if there ever was a bomberman level)
function main()
	-- not used yet since we don't really have any spawning entities
	spawnObjSystemUpdateTags(0,-1)
	GameAPI:initPlayerSpawns()
	local avoid = GameAPI:calcAvoids()
	local breaks = GameAPI:spawnBreaks(avoid)
	GameAPI:addItems(breaks)

--[[
now i wanna add AI players
but we need to wait for all clients to finish spawning ...
and we don't know when that will happen ...
nor will players know if there is an AI associated with them ...

time to redo the clientSpawn system:
1) enumerate all playerstart spawnobjects
2) keep track of an association from playerstart spawnobjects to players/ai's (let this be a lesson to move the spawnobjs all to lua)
- when a player spawns, pick a non-ai'd spot.  if none are left then boot an ai.  if none are left then complain you can't connect
- when an ai spawns find the next unavailable spot
- expose functionality for the following: addAI() removeAI()
--]]

	setTimeout(function() setAI(persist.ai) end, .2)
end

function setAI(num)
	num = castnumber(num)
	if num < 0 then num = 0 end
	-- dumb down
	while num < #GameAPI.aiPlayers do
		local v = GameAPI.aiPlayers[#GameAPI.aiPlayers]
		v.ent:delete()
		v.ent = nil
	end
	-- grow up
	while num > #GameAPI.aiPlayers do
		for i,v in ipairs(GameAPI.playerSpawns) do
			if not v.ent or not v.ent:isActive() or not v.ent:isa(Player) then
				GameAPI.aiPlayers:insert(v)
				-- and spawn an AI for our spot
				v.ent = PlayerAI:new(v.so)
				break
			end
		end
	end

	-- for change levels and what not
	persist.ai = #GameAPI.aiPlayers
end

