require 'bomberman.scripts.base'

entclass('Item', Base)

Item.INVINCIBLE_DURATION = .5	-- for this long after they spawn, items can't be killed

function Item:ctor(so)
	Item.super.ctor(self,so)
	self.spawnTime = getTime()
end

function Item:touch(other)
	Item.super.touch(self,other)
	if other and other:isa(Player) then
		other.items:insert(self)
		if not other[self.itemParam] then
			other[self.itemParam] = 1
		else
			other[self.itemParam] = other[self.itemParam] + 1
		end
		-- hide until they die
		self:setNoPhysics(true)	-- should cut out touches
		self:setAnimGroup('null')
		self:setPos{-10,-10}	-- put it somewhere it wont interfere
	end
end

function Item:hitByBomb()
	if getTime() > self.spawnTime + self.INVINCIBLE_DURATION then
		--disintigrate!
		self:delete()
	end
end

-- called in Player:die as it is getting rid of its items
function Item:respawn()
	-- find a random square far enough from every player (who is sitll alive) .. or every item, i should say ...

	local tries = 1
	local newpos = vec2:new()
	while tries < 100 do
		-- minus a half so we end up [0,n-1]+.5 rather than [1,n]
		newpos[1] = math.random(mapGetWidth())-.5
		newpos[2] = math.random(mapGetHeight())-.5
		if mapGetType(unpack(newpos)) == MAPTYPE.Empty then
			local close = false
			for e in allents() do
				-- todo - if the ent is a breakable then stuff the item back inside of it?
				if (newpos - e:getPos()):lenSq() < 1 then
					close = true
					break
				end
			end
			if not close then break end
		end
		tries = tries + 1
	end
	if tries >= 100 then
		print('couldnt spawn an item =( lost it to the deep ...')
	else
		self:setPos(newpos)
		self:setAnimGroup(self.class.animGroup)
		self:setNoPhysics(false)
	end
end

itemParams = table.new()

local bombermanItemIndex = 1
local function makeItem(className, spriteFileName, itemParamName, itemParamDefaultValue)
	local group = 'bombermanitem'..bombermanItemIndex
	bombermanItemIndex = bombermanItemIndex + 1

	-- keep track of all params - for resetting
	-- keep it in a key/value combo for unification's sake (move somewhere else?)
	itemParamDefaultValue = castnumber(itemParamDefaultValue)
	itemParams[itemParamName] = itemParamDefaultValue

	animLoad(group, spriteFileName)

	return entclass(className, Item, { animGroup = group, itemParam = itemParamName })
end

makeItem('FlameItem', 'flame_item.png', 'flameCount', 1)
makeItem('BombItem', 'bomb_item.png', 'bombCount', 1)
makeItem('IncineratorItem', 'incin_item.png', 'incinCount')
makeItem('RedBombItem', 'redbomb_item.png', 'redbombCount')

