require 'base.scripts.ents'
require 'base.scripts.lighttex'
require 'base.scripts.clientinput'
require 'base.scripts.client'		--createClientStartClass
require 'base.scripts.menuent'		--makeEntityMenuCapable
require 'base.scripts.defs'
require 'base.scripts.hud'

-- here is our mod player
-- its gonna be like the original player except
-- i'm gutting the weapon system
-- and gonna put in more control for scripting
-- and do more with the whole grabbing objects thing ...

animLoad('ninja', {
	sprites = {
		{file='hero/stand.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/climb.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/run1.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/run2.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/run3.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/attack1.png', bbox={-.75,-1.5,.75,.5}};
		{file='hero/attack2.png', bbox={-.75,-1.5,.75,.5}};
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		pain = {{joint=1; time=0; sprite=2}};
		run = {
			{joint=1; time=0; sprite=3};
			{joint=1; time=.1; sprite=4};
			{joint=1; time=.2; sprite=5};
			{joint=1; time=3};
		};
		jump = {{joint=1; time=0; sprite=4}};
		attack = {
			{joint=1; time=0; sprite=6};
			{joint=1; time=.1; sprite=7};
			{joint=1; time=.2};
		};
		die = {{joint=1; time=0; sprite=1}};
	};
})


-- here's my clientKeyXXX(this->in.input_cmd, INKEY_YYY) replacement:
-- self.input.key.YYY.XXX
-- where YYY is (engine -> script)
-- UP		up
-- DOWN		down
-- LEFT		left
-- RIGHT	right
-- ATTACK	attack
-- HOOK		hook
-- WEAPPREV	weapPrev
-- WEAPNEXT	weapNext
-- where XXX is (engine -> script)
-- State	state
-- Up		isUp
-- Down		isDown
-- Press	isPressed
-- Release	isReleased

-- this one is found in the physics module:
local baseFrictionScalar = 70

entclass('Player', TakesDamage,{

	aimAngleRange = 85,
	aimAngleRate = 2,
	wallJumpThreshold = 0.5,
	wallJump = vec2:new(10),
	swimForce = 250,
	jumpVel = 21.5,
	jumpFromWaterVel = 21.5;
	runForce = 15,
	airForce = 15,
	runMaxVel = 20,
	painStun = 0.5,			-- how long you're stunned after pain
	painDelay = 1.5,		-- 
	flashlight = 0,

	jetpackHorzForce = vec2:new(40,40),
	jetpackVertForce = 60,
	jetpackIllum = vec3:new(1),
	jetpackIllumSize = vec2:new(15,4),
	jetpackRandCoeff = 0.7,
	
	illum = vec3:new(1),
	illumWidth = 12,
	illumHeight = 1.5,
	
	disableInput = false,

	updateTags = function(self, userTag)	-- void (entity, int)
		spawnObjSystemUpdateTags(userTag, self.tag)
		self.tag = userTag
	end,
	
	updateTagsFromPos = function(self, pos)	-- void (entity, vec2f)
		local userTag = mapGetTag( math.floor(pos[1]), math.floor(pos[2]) )
		self:updateTags(userTag)
	end,
	
	update = function(self)
		self:updateTagsFromPos(self:getPos())
		if getTime() > self.screenFlashEndTime then
			self.screenFlash:set(0)
		end
	end,
	
	updateViewpos = function(self)
		-- update the viewpos
		if not self.disableInput then
			if self.input.mouse.active then
				local bounds = viewGetBox()
				local halfScreen = bounds:size()
				local mouseFrac = self.input.mouse.pos - .5
				self.destViewpos = vec2:new(self:getPos()) + mouseFrac * halfScreen * 0.8
			end
		end
		self.viewpos = self.viewpos + (self.destViewpos - self.viewpos) * 0.0625
	end;
	
	think = function(self)
		Player.super.think(self)
		if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
		
		local viewbox = viewGetBox()
		local mouseWorldPos = vec2:new(self.input.mouse.pos):map(viewbox)
		
		self:update()
		self:updateViewpos()
		
		if self.dead then return end
		
		if self.input.mouse.active then
			self.aimDir = (mouseWorldPos - self:getPos()):unit()
		else
			self.aimAngle = clamp(self.aimAngle, -self.aimAngleRange, self.aimAngleRange)
			self.aimDir = vec2.expd(self.aimAngle)
			if self:getLeft() then self.aimDir[1] = -self.aimDir[1] end
		end

		-- if we went from inside water to outside water then make the player hop up a block's worth
		local newMapType = mapGetType(self:getPosX(), self:getPosY(), self:getLayer())
		if newMapType ~= self.lastMapType then
			if self.lastMapType == MAPTYPE.Liquid and newMapType ~= MAPTYPE.Liquid then
				if self:getVelY() > -self.jumpFromWaterVel then
					self:setVelY(-self.jumpFromWaterVel)
				end
			end
		end
		self.lastMapType = newMapType
		
		if self.pickup then
			local center = vec2:new(self:getPosX(), self:getPosY() - 1)
			self.pickup:setVel(vec2:new(self:getVel()) + self.aimDir * 7.5)
			self.pickup:setPos(center + self.aimDir * .5)
		end

		if getTime() - self.lastPainTime < self.painStun then return end

		if self.pickup then
			local o = self.pickup
			-- assert(o:isa(Collidable))
			if not self.disableInput and self.input.key.attack.isUp then
				o:setNoPhysics(false)
				if o:getSpawnObj() then
					if o:getSpawnObj()._ent then
						o:delete()
					else
						o:getSpawnObj()._ent = o
					end
				end
				self.pickup = nil
			end
		end
		
		--recent hack
		-- try to cut the wire when the script says so
		if self.disableInput then
			self.input:setcmd(0)
			self.input.mouse.active = false
		end
		
		if self.input.key.left.isPressed then self:setLeft(true) end
		if self.input.key.right.isPressed then self:setLeft(false) end
		
		if not self.input.mouse.active then
			if self.input.key.up.isDown then self.aimAngle = self.aimAngle - self.aimAngleRate end
			if self.input.key.down.isDown then self.aimAngle = self.aimAngle + self.aimAngleRate end
		else
			self:setLeft(mouseWorldPos[1] < self:getPosX())
		end
		
		if self:getContactNormalY() > -0.5 then
			if self:getLeft() and self:getContactNormalX() > 0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			elseif not self:getLeft() and self:getContactNormalX() < -0.5 and
				getTime() - self:getContactTime() < self.wallJumpThreshold and
				getTime() - self.lastWallJumpTime > self.wallJumpThreshold
			then
				self:setAnimSeq('grab')
			else
				self:setAnimSeq('jump')
			end
		else
			if self.attackFinished > getTime() then
				if self.input.key.left.isUp and
					self.input.key.right.isUp and
					self.input.key.up.isUp
				then
					self:setAnimSeq('attack')
				end
			else
				if self.input.key.left.isDown or
					self.input.key.right.isDown
				then
					self:setAnimSeq('run')
				else
					self:setAnimSeq('idle')
				end
			end
		end
		
		-- swimming
		
		if mapGetType(self:getPosX(), self:getPosY(), self:getLayer()) == MAPTYPE.Liquid then
			self:setAnimSeq('jump')	--swim
			local force = math.sin(getTime() * 2 * math.pi) * 0.5 + 0.5
			force = force * self.swimForce * getDeltaTime()
		
			if self.input.key.up.isDown then
				self:setVelY(self:getVelY() - force)
				
				if mapGetType(self:getPosX(), (self:getPosY() - 0.1), self:getLayer()) ~= MAPTYPE.Liquid then
					self:setVelY(-0.5 * self.jumpVel)
				end
			end
			if self.input.key.down.isDown then self:setVelY(self:getVelY() + force) end
			if self.input.key.left.isDown then self:setVelX(self:getVelX() - force) end
			if self.input.key.right.isDown then self:setVelX(self:getVelX() + force) end
			
		-- running
			
		elseif self:getContactNormalY() < -0.7 then
			if self.input.key.up.isDown then
				local n = (vec2:new(self:getContactNormal()) + vec2:new(0,-1)):unit()
				self:setVel(vec2:new(self:getVel()) + n * self.jumpVel)
			end
			if self.input.key.left.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() > -self.runMaxVel then
					self:setVelX(self:getVelX() - (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() < -self.runMaxVel then
						self:setVelX(-self.runMaxVel)
					end
				end
			end
			if self.input.key.right.isDown then
				local grfr = mapGetFriction(self:getContactTileX(), self:getContactTileY(), self:getLayer(), SIDE.UP)
				local coeff = 2 - 2 / (grfr * grfr + 1)
				if self:getVelX() < self.runMaxVel then
					self:setVelX(self:getVelX() + (self.runForce * coeff + baseFrictionScalar * grfr * self:getFriction()) * getDeltaTime())
					if self:getVelX() > self.runMaxVel then
						self:setVelX(self.runMaxVel)
					end
				end
			end
			
		-- flight / wall movement
		
		else

			-- wall movement
			
			if self:getContactNormalX() < -0.7 or self:getContactNormalX() > 0.7 then
				self.lastWallJumpTime = getTime() - self.wallJumpThreshold - 1
			end
			
			if self.input.key.left.isDown then
				if self:getContactNormalX() < -0.5 and
					getTime() - self:getContactTime() < self.wallJumpThreshold and
					getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
				then
					self.lastWallJumpTime = getTime()
					self:setVelX(self:getVelX() - self.wallJump[1])
					self:setVelY(self:getVelY() - self.wallJump[2])
				else
					if self:getVelX() > -self.runMaxVel then
						self:setVelX(self:getVelX() - self.airForce * getDeltaTime())
					end
				end
			end

			if self.input.key.right.isDown then
				if self:getContactNormalX() > 0.5 and
					getTime() - self:getContactTime() < self.wallJumpThreshold and
					getTime() - self.lastWallJumpTime > self.wallJumpThreshold 
				then
					self.lastWallJumpTime = getTime()
					self:setVelX(self:getVelX() + self.wallJump[1])
					self:setVelY(self:getVelY() - self.wallJump[2])
				else
					if self:getVelX() < self.runMaxVel then
						self:setVelX(self:getVelX() + self.airForce * getDeltaTime())
					end
				end
			end
		end				
	end,
	
	touch = function(self,other)
		Player.super.touch(self,other)
		if not other then return end
		if other:getNoPhysics() then return end
		if not other.canBePickedUp then return end	-- here's our new interface for pickupable's
		if self.disableInput then return end
		if self.input.key.attack.isUp then return end
		if self.pickup then return end
		other:setNoPhysics(true)

		if other:getSpawnObj() then
			other:getSpawnObj()._ent = nil
		end

		self.pickup = other
	end,
	
	respawn = function(self)
		-- for single player, reload the map
		-- session vars will pick us up where we left off
		resetSession()
	end,
	
	takeDamage = function(self, other, damage)
		if getTime() - self.lastPainTime < self.painDelay then return end
		Player.super.takeDamage(self, other, damage)
	end,
	
	pain = function(self, other)
		if getTime() - self.lastPainTime < self.painDelay then return end
		
		self.screenFlash:set(1,0,0,.25)
		self.screenFlashEndTime = getTime() + 0.05
		self.lastPainTime = getTime()
		
		self:setVelY(self:getVelY() - 10)
		if other and other:getPosX() < self:getPosX() then
			self:setVelX(self:getVelX() + 10)
		else
			self:setVelX(self:getVelX() - 10)
		end
		
		self:setAnimSeq('pain')
	end,
	
	death = function(self, other)
		if self.dead then return end
		
		self:setLeft(false)
		self:setAngle(-90)
		self:setColor{1,0,0,1}
		self.dead = true
		
		self:setAnimSeq('die')
		-- respawn after the animation sequence
		setTimeout(function() self:respawn() end, 3)
	end,

	init = function(self,pos)
		self:setPos(pos)
		self.viewpos = vec2:new(pos)
		self.destViewpos = vec2:new(pos)
		self:setBBox{min={-.4,-1.4},max={.4,.4}}
		self:setFriction(1)
		self.health = 100
		self.maxHealth = self.health
		self:setColor{1,1,1,1}
		
		self:setAnimGroup('ninja')
		self:setAnimSeq('idle')
		
		self.screenFlash = vec4:new()
		self.screenFlashEndTime = -1
		self.aimAngle = 0
		self.lastPainTime = 0
		self.attackFinished = 0
		self.lastWallJumpTime = 0

		self.jetpackEndTime = -1
		self.flipoutEndTime = -1
		
		-- for engine binding
		-- soon to be abstracted for all entities ...
		self.input = ClientInput:new()
		
		self:updateTags(-1)
	end,
	
	ctor = function(self, table)
		Player.super.ctor(self, table)
		self:init(self:getPos())
	end,
})
function Player.clientUpdate(classObj)
	Player.super.clientUpdate(classObj)

	if net.flashlight ~= 0 then
	
		local lightTex = circularIntenseLightTex
		if net.flashlight == 1 then lightTex = spotLightTex end

		local lightSpot = vec2:new(net.pos[1], net.pos[2] - 1)
		local rad = classObj.illumWidth
		if net.flashlight ~= 1 then rad = (classObj.illumWidth + classObj.illumHeight) * 0.5 end
		local lightRange = box2:new(-rad, rad)
		local lightRot = vec2:new(net.aimDir)
		local lightPos = vec2:new(lightSpot)
		if net.flashlight == 1 then
			self:setPos(net.pos + net.rot * lightRange.max[1])
		end
		lightSpot = lightSpot + lightRot
		
		-- TODO - move to clientUpdate
		addLight(lightPos, classObj.illum, lightTex, lightRange, lightSpot, lightRot)
	end
	
	if net.screenFlash[4] then
		screenFlash(net.screenFlash)
	end
end


-- basically stole from ClientView
function Player:serverRecvMessage(cmd, ...)
	local f = self[cmd]
	if not f then
		print('Player:serverRecvMessage got an unknown command ',cmd)
	else
		f(self, ...)
	end
end
function Player.clientRecvMessage(cmd, ...)
	local cl = ClientSide:getClient()
	local f = cl[cmd]
	if not f then
		print('Player.clientRecvMessage got an unknown command ',cmd)
	else
		f(cl, ...)
	end
end
function Player.clientUpdate(classObj)
	Player.super.clientUpdate(classObj)
	ClientSide:getClient():update()
end


	
Player:createNetTable{
	viewpos = 'vec2f';		-- mandatory for client entities
	health = 'float';
	maxHealth = 'float';
	screenFlash = 'color';
	flashlight = 'int';
}

createClientStartClass(Player, 'ninja', box2:new(-.4,-1.4,.4,.4))	-- make Player spawnable, but not until the client connects

makeEntityMenuCapable(Player)

-- this is all a mess atm
viewTileSize(32)
