require 'base.scripts.ents'

animLoad('thwamp', {
	sprites = {
		{file='thwamp.jpg', bbox={-1,-1.5,1,1.5}};
		{file='thwamp2.jpg', bbox={-1,-1.5,1,1.5}};
		{file='thwamp3.jpg', bbox={-1,-1.5,1,1.5}};
	};
	seqs = {
		idle = {{joint=1; time=0; sprite=1}};
		angry = {{joint=1; time=0; sprite=2}};
		drop = {{joint=1; time=0; sprite=3}};
		die = {{joint=1; time=0; sprite=3}};
	};
})

entclass('Thwamp', TakesDamage, {
	animGroup = 'thwamp';
	bbox = box2:new(-1, -1.5, 1, 1.5);
	STATE_WAITING = {},
	STATE_FALLING = {},
	STATE_RISING = {},
	force = 20,
	damage = 5,
	nearRange = 3,
	farRange = 6,
	maxHealth = 100,

	ctor = function(self, so)
		Thwamp.super.ctor(self, so)
		self.state = self.STATE_WAITING			-- STATE_RAISING just to get them back where they belong?
		self:setUseGravity(false)
		
		if so.force then self.force = tonumber(so.force) end
		if so.damage then self.damage = tonumber(so.damage) end
		if so.nearRange then self.nearRange = tonumber(so.nearRange) end
		if so.farRange then self.farRange = tonumber(so.farRange) end
	end,

	think = function(self)
		Thwamp.super.think(self)
		if not self:isActive() then return end	-- needed for any whose parent think() might kill themselves (Collidable)
		if self:getIgnoreEnts() then return end
		if self.state ~= self.STATE_WAITING then return end
		local getmad = false
		for other in allents() do
			if other:isa(TakesDamage) and not other:isa(Thwamp) then
				local dx = self:getPosX() - other:getPosX()
				if dx > -self.farRange and dx < self.farRange then self.getmad = true end
				if dx > -self.nearRange and dx < self.nearRange then
					self:setAnimSeq('drop')
					self:setUseGravity(true)
					self.state = self.STATE_FALLING
					return
				end
			end
		end
		
		if getmad then 
			self:setAnimSeq('angry')
		end
	end,

	touch = function(self, other)
		Thwamp.super.touch(self,other)
		self:setVelX(0)
		if other then
			if other:isa(TakesDamage) and not other:isa(Thwamp) then
				other:takeDamage(self, self.damage)
				if other:getPosX() < self:getPosX() then
					other:setVelX(other:getVelX() - other:getMassInv() * self.force)
				else
					other:setVelX(other:getVelX() + other:getMassInv() * self.force)
				end
			end
		else
			if self.state == self.STATE_WAITING then
				return	-- need this at all?
			elseif self.state == self.STATE_FALLING then
				self:setUseGravity(false)
				self:setVelY( -1)
				self.state = self.STATE_RISING
				self:setAnimSeq('idle')
			elseif self.state == self.STATE_RISING then
				self:setVelY(0)
				self.state = self.STATE_WAITING
			end
		end
	end,
})
Spawnable(Thwamp)
