require 'ddiad.scripts.tile'

entclass('Tile3x3', Tile)

Tile3x3.nbhdbits = 0	-- default instance, for nettables

function Tile3x3:ctor(so)
	Tile3x3.super.ctor(self,so)
	-- now 'colorObj' is the destination parameter
	-- 3x3's have eight neighborhood constraints as well, with a possible option of nil
	self.nbhdReq = table.new()
	for i=1,3 do
		self.nbhdReq[i] = table.new()
		for j=1,3 do
			local ic = 0
			if math.random() < .3 then
				ic = math.random(colorsUsed)
			end
			if ic > 0 then 	-- if it equals zero then keep it nil.  otherwise . . .
				self.nbhdReq[i][j] = colors[ic]
			end
		end
	end
end

function Tile3x3:canPlay(place)
	-- get the location on the board of the place
	if not place.isABoardPlace then	-- if its not a board place then you can only play it
		--return not place.tile		-- ... only if there's not already something there ... right? wrong! because right now playing a piece just replaces its dest with our set color ...  there is no option for setting what to where.  we need a separate function for that ... ?
		return false
	end
	-- place.boardPos is the pos on the board where place is located
	-- whose board you ask?  why place.board of course!
	for i=1,3 do
		for j=1,3 do
			if self.nbhdReq[i][j] then
				-- calc the board pos
				local boardPos = place.boardPos + (vec2:new(i,j) - vec2:new(2))
				
				local board = place.board
				if board.isToric then
					boardPos[1] = (boardPos[1] - 1) % board.size + 1
					boardPos[2] = (boardPos[2] - 1) % board.size + 1
				end
				
				if boardPos[1] >= 1 and boardPos[2] >= 1 and
					boardPos[1] <= place.board.size and boardPos[2] <= place.board.size
				then
					local boardTile = board.places[boardPos[1]][boardPos[2]].tile
					-- if there is a tile there then
					if boardTile then 
						-- if its color doesn't match then bail
						if self.nbhdReq[i][j] ~= boardTile.colorObj then return false end
					end
				end
			end
		end
	end
	-- all systems go!
	return true
end
	
-- update our network vars
function Tile3x3:think()
	Tile3x3.super.think(self)
	if not self:isActive() then return end
	
	local bitindex = 0
	local iv = 0	-- i hope lua numbers can hold this kind of precision . . .
	for i=1,3 do
		for j=1,3 do
			local nbits = 0
			if self.nbhdReq[i][j] then nbits = self.nbhdReq[i][j].index end
			iv = bitor(iv, bitshiftleft(nbits,bitindex))
			bitindex = bitindex + 3
		end
	end
	self.nbhdbits = iv			-- store in case we are run locally
end

-- need to work on the inheritence of static methods on the clientside
function Tile3x3.renderTile(classObj, b)
	Tile3x3.super.renderTile(classObj, b)
	local nbhdbits = net.nbhdbits
	if nbhdbits ~= 0
	-- and nbhdbits ~= bitand(bitshiftleft(color, 15), 7)
	then 
		drawRect{
			color=vec4:new();
			box=box2:new(.2,.8):map(b);
			depth=net.depth-.1;
			tex=buttonMaskTex;
			blend=BLEND.DECAL;
		}
		
		-- now draw some sub-boxes on top of it too to show the constraint arrangement and transition (the rule)
		local bitindex = 0
		for i=1,3 do
			for j=1,3 do
				local nbits = bitand(bitshiftright(nbhdbits, bitindex), 7)
				bitindex = bitindex + 3
				if nbits > 0 and nbits <= colorsUsed then
					classObj.drawButton{
						color=colors[nbits].color;
						box=box2:new(.1,.9):addTo(i,j):divTo(5):map(b);
						depth=net.depth-.2;
					}
				elseif nbits ~= 0 then
					print('Tile3x3.renderTile got a bad neighbor value '..nbits..' for bit index '..bitindex)
				end
			end
		end
	end
end

Tile3x3:createNetTable{
	nbhdbits = 'uint';	-- each neighbor takes 3 bits to represent, so 3*9 = 27 bits, while our int has 32 to share
}
