/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include <string>

//getcwd.  todo - unify in base or system or something
#ifdef WIN32	//MSVC only
#include <windows.h>
#include <direct.h>
#else
#include <sys/unistd.h>
#endif

#include "system/exception.h"
#include "system/system.h"
#include "base/base.h"

#include "inspiration/maptex.h"

#include "impl.h"
#include "func.h"
#include "wndtex.h"
#include "menulib.h"

using namespace std;
using namespace pwn;

//*********************
//TEXTURE EDITOR WINDOW
//*********************

// ************************* MenuTexEdit prototype ************************* 

class MenuTexEdit : public MenuEditorWindow {
	typedef MenuEditorWindow super;
protected:
	EditorImpl *ed;
public:
	int curTexIndex;
	bool doDestroyOnCancel;
	
	RadioGroup rg;
	
	gc::Pointer<Radio> radio2D;
	gc::Pointer<TextField> editTex2D;
	gc::Pointer<Text> buttonOpen2D;
	
	gc::Pointer<Radio> radioCube;
	std::vector<gc::Pointer<TextField> > editTexCube;	//ft, bk, lf, rt, up, dn
	std::vector<gc::Pointer<Text> > buttonOpenCube;
	
	gc::Pointer<Checkbox> checkCycle;
	gc::Pointer<Select> selectNextTex;
	
	gc::Pointer<TextField> editDelay;
	
	gc::Pointer<TextField> editName;
	gc::Pointer<TextField> subX, subY, subW, subH;

	gc::Pointer<Text> buttonOK, buttonCancel;
	
	MenuTexEdit(EditorImpl *);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};


class MenuTex;
class MenuSelectText : public Select {
	typedef Select super;
protected:
	MenuTex *texWnd;
public:
	MenuSelectText(MenuTex *texWnd_, const vec2f &_pos, float _width) : texWnd(texWnd_), super(_pos, _width) {}
	virtual void userSetSelected(int _sel);
};


class MenuTex : public MenuEditorWindow {
	typedef MenuEditorWindow super;
protected:
	EditorImpl *ed;
public:

	gc::Pointer<Text> buttonCreate, buttonModify, buttonDelete;
	gc::Pointer<MenuSelectText> selectTex;
	
	MenuTex(EditorImpl *);
	virtual eventResponse_t mouseEvent(MouseEvent e, const vec2f &v);
};


class EditorWndTexImpl {
protected:
	EditorImpl *ed;
public:
	gc::GlobalPointer<MenuTex> texWnd;
	gc::GlobalPointer<MenuTexEdit> texEditWnd;

	EditorWndTexImpl(EditorImpl *ed_)
	:	texWnd(gc::sys->createPtr<MenuTex>()),
		texEditWnd(gc::sys->createPtr<MenuTexEdit>()),
		ed(ed_)
	{}
	
	void init();
	void popup(int tex_index, bool destroy_on_cancel = false);
	void refresh();
	void refreshAllTextures();
};

	
MenuTexEdit::MenuTexEdit(EditorImpl *ed_) 
:	super(vec2f(1,1), vec2f(40,24)),
	radio2D(createPtr<Radio>()),
	editTex2D(createPtr<TextField>()),
	buttonOpen2D(createPtr<Text>()),
	radioCube(createPtr<Radio>()),
	checkCycle(createPtr<Checkbox>()),
	selectNextTex(createPtr<Select>()),
	editDelay(createPtr<TextField>()),
	editName(createPtr<TextField>()),
	subX(createPtr<TextField>()),
	subY(createPtr<TextField>()),
	subW(createPtr<TextField>()),
	subH(createPtr<TextField>()), 
	buttonOK(createPtr<Text>()),
	buttonCancel(createPtr<Text>()),
	curTexIndex(-1),
	doDestroyOnCancel(false),
	ed(ed_)
{

	visible = false;
	allowHide = false;	//only let the user click 'cancel' - so all is routed through that handler - for the doDestroyOnCancel's sake

	addChild(radio2D = new Radio(vec2f(1,1), "2D", &rg));
	addChild(editTex2D = new TextField(vec2f(5,3), 30));
	addChild(buttonOpen2D = new Text(vec2f(36,3), ".."));
	
	const char *abbrev[] = { "ft", "bk", "up", "dn", "rt", "lf" };
	addChild(radioCube = new Radio(vec2f(1,5), "Cube", &rg));
	for (int i = 0; i < 6; i++) {
		addChild(new Text(vec2f(2,6+i), abbrev[i]));
		TextField *tf = new TextField(vec2f(5,6+i), 30);
		editTexCube.push_back(createPtr<TextField>(tf));
		addChild(tf);
		Text *t = new Text(vec2f(36,6+i), "..");
		buttonOpenCube.push_back(createPtr<Text>(t));
		addChild(t);
	}
	
	addChild(checkCycle = new Checkbox(vec2f(1,13), "Cycle to"));
	addChild(selectNextTex = new Select(vec2f(10,13), 30));
	addChild(new Text(vec2f(2,14), "after"));
	addChild(editDelay = new TextField(vec2f(8,14), 10));
	addChild(new Text(vec2f(20,14), "seconds"));
	
	addChild(new Text(vec2f(1,16), "Name:"));
	addChild(editName = new TextField(vec2f(7,16), 30));
	
	addChild(new Text(vec2f(1,18), "Subtexture:"));
	addChild(new Text(vec2f(2,19), "X:"));
	addChild(subX = new TextField(vec2f(4,19),6));
	addChild(new Text(vec2f(2,20), "Y:"));
	addChild(subY = new TextField(vec2f(4,20),6));
	addChild(new Text(vec2f(11,19), "W:"));
	addChild(subW = new TextField(vec2f(13,19),6));
	addChild(new Text(vec2f(11,20), "H:"));
	addChild(subH = new TextField(vec2f(13,20),6));

	addChild(buttonOK = new Text(vec2f(7,22), "OK"));
	addChild(buttonCancel = new Text(vec2f(10,22), "Cancel"));
}

eventResponse_t MenuTexEdit::mouseEvent(MouseEvent e, const vec2f &v) {
	//handle buttons here (Text)
	if (e.lpress) {
		if (selectedChild.ptr() == buttonOK) {			
			char buffer[7][1024];
			const char *buffers[7] = {		//cubemap uploads
				buffer[0],
				buffer[1],
				buffer[2],
				buffer[3],
				buffer[4],
				buffer[5],
				buffer[6]
			};

			//first check the name for validity:
			strncpy(buffer[0], editName->getText().c_str(), sizeof(buffer[0]));
			if (!editor->nameIsValid(buffer[0])) {
				//TODO - popup?
				con << "Found an invalid texture name\n";
				return RESPONSE_STOP;
			}
			//get the old name
			//if the two differ
			MapTex *mt = ed->world->texs->texs[curTexIndex];
			if (strcmp(buffer[0], mt->texName.c_str())) {
				//then not only update this one . . .
				mt->texName = buffer[0];
				//but also, refresh all the texture lists and such
				editor->refreshAllTextures();
			}

			if (radio2D->isChecked()) {
				//2d

				//this code matches the Refresh_WorldTexList() code.
				//maybe the two should be combined into one method huh?
				if (mt->subtexSize.x > 0 && mt->subtexSize.y > 0) {
					con.print("creating 2d subtex\n");
					ed->world->texs->set2DSub(curTexIndex, editTex2D->getText().c_str(), VEC2ELEM(mt->subtexOffset), VEC2ELEM(mt->subtexSize));
				} else {
					con.print("creating 2d tex\n");
					ed->world->texs->set2D(curTexIndex, editTex2D->getText().c_str());
				}
			} else {
				//cube
				//the order of things: "ft", "bk", "up", "dn", "rt", "lf"
				for (int i = 0; i < 6; i++) {
					strncpy(&buffer[i][0], editTexCube[i]->getText().c_str(), sizeof(buffer[0]));
				}
				con.print("creating cube tex\n");
				ed->world->texs->setCube(curTexIndex, buffers);
			}
			//...and set the texture animation stats / type stats
			if (!checkCycle->isChecked()) {
				mt->nextIndex = -1;		//-1 == no animation
				mt->duration = -1;	//<0 == no animation
			} else {
				mt->nextIndex = selectNextTex->selectedIndex;
				mt->duration = (float)atof(editDelay->getText().c_str());
			}
			
			visible = false;
			//stop should tell the menu to stop input from hitting the outside world
			//but we still get one tile click in the editor ...
			//hmm ... maybe its cuz our window is closing?
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == buttonCancel) {
			con << "clicking button cancel\n";

			if (doDestroyOnCancel) {
				ed->world->texs->destroy(curTexIndex);
			}
			visible = false;
			return RESPONSE_STOP;
		}

		class tex_open_pair_t {
		public:
			Text *button;
			TextField *text;
		};
	
		tex_open_pair_t open_wins[] = {
			{buttonOpen2D,		editTex2D},
			{buttonOpenCube[0],	editTexCube[0]},
			{buttonOpenCube[1],	editTexCube[1]},
			{buttonOpenCube[2],	editTexCube[2]},
			{buttonOpenCube[3],	editTexCube[3]},
			{buttonOpenCube[4],	editTexCube[4]},
			{buttonOpenCube[5],	editTexCube[5]},
		};
		int i;
		tex_open_pair_t *pair;
		for (i = 0, pair = open_wins; i < numberof(open_wins); i++, pair++) {
			if (selectedChild.ptr() != pair->button) continue;
			// do an open window dialog box
			static char dest_filename[1024] = {0};
	
	//			if (dest_filename[0] == 0)		//if it hasnt been init'd yet...
	//				memcpy(dest_filename, texture_filename, MAX_FN_LEN);
#if 0 //def WIN32
			OPENFILENAME o = {
				sizeof(OPENFILENAME),				//length, in bytes, of structure
				main_display->displayWnd,			//handle to window owner
				GetModuleHandle(NULL),				//instance
				"*.bmp; *.tga\0*.bmp; *.tga\0\0",	//filter text
				NULL,								//custom filter
				0,									//custom filter size
				1,									//default filter to use
				dest_filename,						//file name to recieve from open dialog
				sizeof(dest_filename),				//file name size
				NULL,								//file title (no pathname)
				0,									//file title size
				NULL,								//initial directory
				NULL,								//dialog title
				OFN_ALLOWMULTISELECT | OFN_EXPLORER |
				OFN_FILEMUSTEXIST,					//flags
				0,									//file name offset
				0,									//file extension offset
				NULL,								//default file extension
				0,									//custom data
				NULL,								//hook
				NULL};								//template name
	
	
			//store the current dir
			//TODO - store this once and always go back to it
			char cwd[256];
			getcwd(cwd, sizeof(cwd));
	
			strcpy(dest_filename, cwd);
			strcat(dest_filename, "\\.");
	
			BOOL success = GetOpenFileName(&o);
			chdir(cwd);
	
			if (success) {
	
				int cwdlen = strlen(cwd);
				if (memcmp(dest_filename, cwd, cwdlen)) {
					Error("File must exist within the base game directory\r\ndest = %s\r\ncur = %s",
						dest_filename, cwd);
				} else {
					char *subfn = dest_filename + cwdlen + 1;	//pass the '/' at the dir end
					pair->text->text = subfn;
				}
			}
#endif
			return RESPONSE_STOP;
		}
	}
	
	super::mouseEvent(e, v);
	return RESPONSE_STOP;
}

void EditorWndTexImpl::popup(int tex_index, bool destroy_on_cancel) {
	// // // //
	texEditWnd->curTexIndex = tex_index;
	texEditWnd->doDestroyOnCancel = destroy_on_cancel;

	MapTex *mt = ed->world->texs->texs[texEditWnd->curTexIndex];
	
	//do the init here:
			//fill the initial values
	if (!mt->isCube) {
		texEditWnd->radio2D->setChecked(true);
		//2d
		texEditWnd->editTex2D->setText(mt->sources[0]);
		for (int i = 0; i < 6; i++) {
			texEditWnd->editTexCube[i]->setText("");
		}
	} else {
		texEditWnd->radioCube->setChecked(true);
		//cube
		texEditWnd->editTex2D->setText("");
		for (int i = 0; i < numberof(mt->sources); i++) {
			texEditWnd->editTexCube[i]->setText(mt->sources[i]);
		}
	}

	texEditWnd->editName->setText(mt->texName);

	//unify me
	const char *ap(const char *, ...);
	texEditWnd->subX->setText(ap("%d", mt->subtexOffset.x));
	texEditWnd->subY->setText(ap("%d", mt->subtexOffset.y));
	texEditWnd->subW->setText(ap("%d", mt->subtexSize.x));
	texEditWnd->subH->setText(ap("%d", mt->subtexSize.y));

	bool isanim = (mt->nextIndex >= 0 && mt->nextIndex < ed->world->texs->texs.size());
	texEditWnd->checkCycle->setChecked(isanim);

	editor->refreshTextureSelect(texEditWnd->selectNextTex);

	if (isanim) {
		char buffer[32];
		texEditWnd->selectNextTex->setSelected(mt->nextIndex);
		sprintf(buffer, "%f", mt->duration);
		editor->trimFloatString(buffer, sizeof(buffer));
		texEditWnd->editDelay->setText(buffer);
	} else {
		texEditWnd->selectNextTex->setSelected(-1);
		texEditWnd->editDelay->setText("");
	}
	
	texEditWnd->setVisible(true);
}


void EditorWndTexImpl::refresh() {
	//update texture list
	refreshAllTextures();
}

void EditorWndTexImpl::refreshAllTextures() {
	editor->refreshTextureSelect(texWnd->selectTex);
	texWnd->selectTex->setSelected(editor->tex);
}

MenuTex::MenuTex(EditorImpl *ed_)
:	super(vec2f(1,1),vec2f(22,5)),
	buttonCreate(createPtr<Text>()),
	buttonModify(createPtr<Text>()),
	buttonDelete(createPtr<Text>()),
	selectTex(createPtr<MenuSelectText>()),
	ed(ed_)
{
	visible = false;
	
	addChild(selectTex = new MenuSelectText(this, vec2f(1,1), 20));
	addChild(buttonCreate = new Text(vec2f(1,3), "Create"));
	addChild(buttonModify = new Text(vec2f(8,3), "Modify"));
	addChild(buttonDelete = new Text(vec2f(15,3), "Delete"));
}

eventResponse_t MenuTex::mouseEvent(MouseEvent e, const vec2f &v) {
	
	//handle buttons here (Text)
	if (e.lpress) {
		if (selectedChild.ptr() == buttonCreate) {
			int index = ed->world->texs->create();
			if (index == -1) return RESPONSE_STOP;
			
			ed->texWnd->popup(index, true);	//true means destroy on cancel
			//set the new guy as the selected
			editor->tex = index;
			//refresh the list
			//this is auto-handled within the dlg proc for windows
			editor->refreshAllTextures();
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == buttonModify) {
			ed->texWnd->popup(editor->tex);
			return RESPONSE_STOP;
		} else if (selectedChild.ptr() == buttonDelete) {
			if (ed->world->texs->destroy(editor->tex)) {
				//keep it inbounds
				int count = ed->world->texs->texs.size();
				if (editor->tex >= count && count > 0) {
					editor->tex = count - 1;
				}
				//refresh /all/ the texture references around the house
				editor->refreshAllTextures();
			}
			return RESPONSE_STOP;
		}
	}
	
	super::mouseEvent(e, v);
	return RESPONSE_STOP;
}

void MenuSelectText::userSetSelected(int _sel) {
	if (_sel == selectedIndex) return;
	super::userSetSelected(_sel);
	if (this == texWnd->selectTex) {
		editor->tex = selectedIndex;
	}
}

void EditorWndTexImpl::init() {
	(*editor->editWnd[EW_TEX]) = texWnd = new MenuTex(ed);
	if (!texWnd) throw Exception() << "EditorWndTex::init failed to create texWnd";
	editor->rootmenu->addChild(texWnd);

	texEditWnd = new MenuTexEdit(ed);
	if (!texEditWnd) throw Exception() << "EditorWndTex::init failed to create texEditWnd";
	editor->rootmenu->addChild(texEditWnd);
}

EditorWndTex::EditorWndTex(EditorImpl *ed) : impl(new EditorWndTexImpl(ed)) {}
EditorWndTex::~EditorWndTex() { delete impl; }
void EditorWndTex::init() { impl->init(); }
void EditorWndTex::refresh() { impl->refresh(); }
void EditorWndTex::refreshAllTextures() { impl->refreshAllTextures(); }

