/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "entconn.h"
#include "entalloc.h"
#include "game/main.h"
#include "game/client.h"

void ServerConnSys::clear(Connection *c) {
	if (c->sender) delete c->sender;
	memset(c, 0, sizeof(*c));
}

void ServerConnSys::clearAll() {
	for (Connection *c = serverConnSys.connections; c < endof(serverConnSys.connections); c++) {
		clear(c);
	}
}

bool ServerConnSys::init() {
	clearAll();
	return true;
}

void ServerConnSys::shutdown() {

	for (Connection *c = serverConnSys.connections; c < endof(serverConnSys.connections); c++) {
		if (c->active) destroy(c);
	}

	//and zero it all
	clearAll();
}

Connection *ServerConnSys::create() {
	for (int i = 0; i < numberof(connections); i++) {
		Connection *c = connections + i;
		if (!c->active) {
			clear(c);

			c->active = true;
#ifndef ONE_DELTA_ENTITY_BUFFER
			c->cl_entity = entAllocSys.getClientEntityArray() + i * sizeof(ClientEntity) * CL_ENTITY_NUM_PER_CLIENT;
#endif
			return c;
		}
	}
	return NULL;
}

void ServerConnSys::destroy(Connection *c) {
	con << "ServerConnSys::destroy: deleting entities...\n";

	//disconnect the entity from the game
	//this assertion is failing...
	//that means this is being called after the server has already shut the game down
	for (int i = 0; i < c->count; i++) {
		//separate on disconnect function anyone?
		game.deleteEntity(c->local[i].ent);
	}

	c->active = false;
}

ServerConnSys serverConnSys;
