/*
	Copyright (c) 2011 Christopher E. Moore ( christopher.e.moore@gmail.com / http://christopheremoore.net )

	Permission is hereby granted, free of charge, to any person obtaining a copy
	of this software and associated documentation files (the "Software"), to deal
	in the Software without restriction, including without limitation the rights
	to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
	copies of the Software, and to permit persons to whom the Software is
	furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in
	all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
	IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
	AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
	LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
	THE SOFTWARE.
*/
#include "sound.h"

#define SOUND_NONE		0	//"I don't know, you tell me"
#define SOUND_OPENAL	1	//file seek error in openal32.lib
#define SOUND_AMP11		2	//working, but stutters when filling the buffer...(mp3 too compressed?)
#define SOUND_FMOD		3

#define SOUND_METHOD	SOUND_NONE

#if SOUND_METHOD == SOUND_NONE

bool soundSystemInit() { return true; }
void soundSystemShutdown() {}
sound_buffer_t soundLoad(const char *filename) {
	sound_buffer_t buffer = {0};
	return buffer;
} //(sound_buffer_t){-1}; }
void soundUnload(sound_buffer_t handle) {}
sound_source_t soundCreate(sound_buffer_t buffer) {
	sound_source_t source = {0};
	return source;
} //(sound_source_t){-1}; }
void soundPlay(sound_source_t handle) {}
void soundPause(sound_source_t handle) {}
void soundStop(sound_source_t handle) {}
float soundGetPos(sound_source_t handle) { return 0; }
float soundGetLen(sound_buffer_t handle) { return 1; }

#elif SOUND_METHOD == SOUND_OPENAL

//doesn't like the openal32 library...
#pragma comment(lib, "openal32.lib")
#pragma comment(lib, "alut.lib")

#include <al\alc.h>
#include <al\alut.h>
#include "../../m_foundation/main.h"

//system

bool soundSystemInit() {
	ALCdevice *device = alcOpenDevice((ALubyte*)"DirectSound3D");
	if (!device) {
		main_console->Print("Failed to Initialize Open AL\n");
		return false;
	}
	//Create context(s)
	ALCcontext *context = alcCreateContext(device,NULL);
	//Set active context
	alcMakeContextCurrent(context);

	// Clear Error Code
	ALint	error;
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("failed to create a context!\n");
		return false;
	}

	// Set Listener attributes

	ALfloat listenerPos[]={0.0,0.0,0.0};
	ALfloat listenerVel[]={0.0,0.0,0.0};
	ALfloat	listenerOri[]={0.0,0.0,-1.0, 0.0,1.0,0.0};	// Listener facing into the screen

	// Position ...
	alListenerfv(AL_POSITION,listenerPos);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alListenerfv POSITION : %d\n", error);
		return false;
	}

	// Velocity ...
	alListenerfv(AL_VELOCITY,listenerVel);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alListenerfv VELOCITY : %d\n", error);
		return false;
	}

	// Orientation ...
	alListenerfv(AL_ORIENTATION,listenerOri);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alListenerfv ORIENTATION : %d\n", error);
		return false;
	}

	return true;
}

void soundSystemShutdown() {
	//Get active context
	ALCcontext *context=alcGetCurrentContext();
	//Get device for active context
	ALCdevice *device = alcGetContextsDevice(context);

	//Disable context
	alcMakeContextCurrent(NULL);
	//Release context(s)
	alcDestroyContext(context);
	//Close device
	alcCloseDevice(device);
}

sound_buffer_t soundLoad(const char *filename) {
	
	sound_buffer_t handle = {0};

	main_console->Print("soundLoad %s\n", filename);

	alGetError();

	// Generate Buffers
	ALint	error;
	alGenBuffers(1, &handle.i);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alGenBuffers : %d\n", error);
		handle.i = 0;
		return handle;
	}

	ALenum	format;
	ALvoid	*data;
	ALsizei size;
	ALsizei freq;
	ALboolean loop;

	//why doesn't alutLoadWAVFile use const char * for its filename?
	alutLoadWAVFile((char *)filename,&format,&data,&size,&freq,&loop);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alutLoadWAVFile : %d\n", error);
		// Delete Buffers
		alDeleteBuffers(1, &handle.i);
		handle.i = 0;
		return handle;
	}

	// Copy footsteps.wav data into AL Buffer 0
	alBufferData(handle.i,format,data,size,freq);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alBufferData buffer 0 : %d\n", error);
		// Delete buffers
		alDeleteBuffers(1, &handle.i);
		handle.i = 0;
		return handle;
	}

	// Unload footsteps.wav
	alutUnloadWAV(format,data,size,freq);
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alutUnloadWAV : %d\n", error);
		// Delete buffers
		alDeleteBuffers(1, &handle.i);
		handle.i = 0;
		return handle;
	}

	return handle;
}

void soundUnload(sound_buffer_t handle) {
	alGetError();
	alDeleteBuffers(1, &handle.i);
	ALint	error;
	if ((error = alGetError()) != AL_NO_ERROR) {
		main_console->Print("alDeleteBuffers : %d\n", error);
		return;
	}
}

sound_source_t soundCreate(sound_buffer_t buffer) {
	alGetError();
	sound_source_t source = {0};
	alGenSources(1,&source.i);
	ALint error;
	if ((error = alGetError()) != AL_NO_ERROR)
	{
		main_console->Print("alGenSources 2 : %d\n", error);
		return source;
	}

	ALfloat source0Pos[]={ 0.0, 0.0, 0.0};	// Front and right of the listener
	ALfloat source0Vel[]={ 0.0, 0.0, 0.0};

	alSourcef(source.i,AL_PITCH,1.0f);
	alSourcef(source.i,AL_GAIN,1.0f);
	alSourcefv(source.i,AL_POSITION,source0Pos);
	alSourcefv(source.i,AL_VELOCITY,source0Vel);
	alSourcei(source.i,AL_LOOPING,AL_FALSE);

	alSourceStop(source.i);
	alSourcei(source.i, AL_BUFFER, buffer.i);

	return source;
}

void soundPlay(sound_source_t handle) {
	alSourcePlay(handle.i);
}

void soundPause(sound_source_t handle) {
	alSourcePause(handle.i);
}

void soundStop(sound_source_t handle) {
	alSourceStop(handle.i);
}

//no compatibility?
float soundGetPos(int handle) {return 1;}
float soundGetLen(int handle) {return 1;}

#elif SOUND_METHOD == SOUND_AMP11

#pragma comment(lib, "amp11lib.lib")

#include <windows.h>
#include <amp11lib/amp11lib.h>

#include "../../m_foundation/main.h"

//the main player
static ALhandle hPlayer = NULL;

bool soundSystemInit() {

	//init lib
	alInitLibrary();

	//choose a device
	int iDevice = -1;		//-1 implies 'default'
	const float fBufferLen = 1.f;	//size of buffer (in seconds)

	hPlayer = alOpenPlayer(iDevice, 44100, ALtrue, fBufferLen);
	if (!hPlayer) {
		main_console->Print("cannot open wave output (do you have winamp running? :) )");
		return false;
	}

	char strDevice[256];
	alDescribePlayerDevice(iDevice, strDevice, sizeof(strDevice));
	main_console->Print("opened output device: %s\n", strDevice);

	const float fTimerFreq = 1.f;	//timer frequency (???)
	alEnableRedirection(fTimerFreq);

	return true;
}

sound_buffer_t soundLoad(const char *filename) {
	sound_buffer_t hFile = {0};
	if (!filename) return hFile;
	if (!filename[0]) return hFile;

	main_console->Print("loading sound %s ...\n", filename);

	hFile.i = alOpenInputFile(filename);
	if (!hFile.i) {
		main_console->Print("failed to load sound %s\n", filename);
	}

	return hFile;
}

void soundUnload(int handle) {
	if (!handle) return;
	alClose(handle);
}


sound_source_t soundCreate(sound_buffer_t buffer) {
	sound_source_t hDecoder;
	hDecoder.i = alOpenDecoder(buffer.i);
	if (!hDecoder.i) {
		main_console->Print("failed to create a decoder\n");
	}

	return hDecoder;
}

void soundPlay(sound_source_t hDecoder) {
	if (!hDecoder.i) return;
//THIS IS CRASHING ON MY COMP! WHATS THE DEAL?
//only during startup.. .for some reason...
	alSetRedirection(hDecoder.i, hPlayer);
}

void soundPause(sound_source_t hDecoder) {
	if (!hDecoder.i) return;
	alSetRedirection(hDecoder.i, 0);
}

void soundStop(sound_source_t hDecoder) {
	if (!hDecoder.i) return;
	alSetRedirection(hDecoder.i, 0);
	alDecSeekAbs(hDecoder.i, 0.f);
}

void soundSystemShutdown() {
	alEndLibrary();
}

float soundGetPos(sound_source_t handle) {
	if (!handle.i) return 0.f;
	return alDecGetPos(handle.i);
}

float soundGetLen(sound_buffer_t handle) {
	if (!handle.i) return 1.f;
	return alDecGetLen(handle.i);
}

#elif SOUND_METHOD == SOUND_FMOD

//system
bool soundSystemInit() {
}

void soundSystemShutdown() {
}

//sound load/unload
int soundLoad(const char *filename) {
}

void soundUnload(int handle) {
}

//handle control (vcr-style)
void soundPlay(int handle) {
}

void soundPause(int handle) {
}

void soundStop(int handle) {
}

float soundGetPos(int handle) {
}

float soundGetLen(int handle) {
}

#endif //SOUND_METHOD


class SoundSys : public IModule {
public:
	virtual bool sysInit() { return soundSystemInit(); }
	virtual void sysKill() { soundSystemShutdown(); }
};
static SoundSys g_soundSys;
IModule &soundSys = g_soundSys;
