require 'base.scripts.util'
require 'topview.scripts.clientview'	-- I'm going to append a lot of methods onto TopviewClientView, so no circular require's back to "battle" please
require 'topview.scripts.battlecmd'
require 'topview.scripts.itemdb'
require 'topview.scripts.spelldb'

gaugeBorderTex = spriteLoad('gaugeBorder.png')

-- depends on battlecmd.lua for battlecmds

	-- BATTLE CMD PICK TARGETS

-- to be called from a thread spawned from the battle menu callback
function battleCmdPickTargets(cmd, battle, team, char, targetting)

	-- figure out the initial target

	local target = char
	if bitand(targetting, targetFlags.targetEnemies) ~= 0 then
		-- pick the first enemy you find
		target = nil
		for _, otherTeam in pairs(battle.teams) do
			if otherTeam ~= team then
				for _,e in ipairs(otherTeam.ents) do
					target = e
					break
				end
			end
			if target then break end
		end
		assert(target)
	end

	-- now deduce what options the user has

	local targetAll = false
	local targetGroup = false
	local targetTeam = false
	
	local targetmethod = bitand(targetting, targetFlags.allOrTeam + targetFlags.groupOrTeam)
	if targetmethod == 0 then 
		-- unit -- dont change anything
	elseif targetmethod == targetFlags.allOrTeam then
		-- all -- add everyone
		targetAll = true
	elseif targetmethod == targetFlags.groupOrTeam then
		-- group -- add everyone in the current 'targets' set's group
		targetGroup = true
	else	-- targetmethod == targetFlags.allOrTeam + targetFlags.groupOrTeam
		-- team -- add everyone in the current 'targets' set's team
		targetTeam = true
	end
	
	-- from the init target and target flags, let the user pick
	
	if bitand(targetting, targetFlags.noPrompt) == 0 then
		local args = {
			class='PickTargetMenu';
			targetID=target._ptr;
			targetGroup=targetGroup;
			targetTeam=targetTeam;
			targetAll=targetAll;
			canToggleGroup = bitand(targetting, targetFlags.toggleGroup) ~= 0;
		}
		local entInfoIndex = 1	-- stagger tables so no one piece of the message is too big
		for teamIndex, otherTeam in pairs(battle.teams) do
			for _,e in ipairs(otherTeam.ents) do
				-- pass info relative to e
				-- index, bbox
				-- TODO - selection buffer anyone?
				local wbox = box2:new(
					e:getBBoxMinX() + e:getPosX(),
					e:getBBoxMinY() + e:getPosY(),
					e:getBBoxMaxX() + e:getPosX(),
					e:getBBoxMaxY() + e:getPosY())
				args['entInfo'..entInfoIndex] = "{'" .. table.new{e._ptr, wbox.min[1], wbox.min[2], wbox.max[1], wbox.max[2]}:concat("','") .. "'}"
				entInfoIndex = entInfoIndex + 1
			end
		end

		--team.client:serverSendMessage('clientBattlePickTarget', targetting, unpack(entargs))
		team.pickTargetResult = nil
		team.client:createPromptingMenu(args)
		target = team.pickTargetResult
		targetAll = team.pickTargetAll
		targetGroup = team.pickTargetGroup
		targetTeam = team.pickTargetTeam
	end
	
	-- go from what the user picked to an entity set

	local targets = table.new()
	if targetAll then
		for _,otherTeam in pairs(battle.teams) do
			for _,e in ipairs(otherTeam.ents) do
				targets:insert(e)
			end
		end
	elseif targetGroup or targetTeam then
		for _,e in ipairs(target.team.ents) do
			targets:insert(e)
		end
	else
		targets:insert(target)
	end
	return targets
end

-- TODO - when creating this make sure you give it the right parent
class('PickTargetMenu', Menu)
PickTargetMenu.pos = {0,0}
PickTargetMenu.size = {unpack(Menu.sysSize)}
PickTargetMenu.backgroundColor = {0,0,0,0}
PickTargetMenu.hasPrompts = true
function PickTargetMenu:onOk(input)
	local mouseWorldPos = vec2:new(input.mouse.pos):map(viewGetBox())
	entClientSendMessage(self.client.entID, 'serverBattlePickTarget', 'pos', unpack(mouseWorldPos))
end
function PickTargetMenu:onCancel()
	self:delete()
end
function PickTargetMenu:onDelete()
	self.client.battle.pickTargetMenu = nil
end
function PickTargetMenu:ctor(args)
	args.size = {unpack(Menu.sysSize)}
	PickTargetMenu.super.ctor(self, args)

	local targetID=args.targetID;
	local targetGroup=args.targetGroup;
	local targetTeam=args.targetTeam;
	local targetAll=args.targetAll;
	local canToggleGroup = bitand(targetting, targetFlags.toggleGroup) ~= 0;
	
	-- to correctly allow closing
	self.client.battle.pickTargetMenu = self
	
	-- now, for all enemies in the battle (sv->cl comm.)
	--	put submenus that accept prompts ... ?  or do like the FF6J does
	-- and put a listing somewhere else...
	local battleEnts = table.new()
	local i = 1
	while true do
		local entInfo = args['entInfo'..i]
		if not entInfo then break end
		local entID = tonumber(entInfo[1])
		local wbox = box2:new(
			tonumber(entInfo[2]),
			tonumber(entInfo[3]),
			tonumber(entInfo[4]),
			tonumber(entInfo[5]))
		local viewbox = viewGetBox()
		local viewsize = viewbox:size()
		local relMin = {
			(wbox.min[1] - viewbox.min[1]) / viewsize[1] * Menu.sysSize[1],
			(wbox.min[2] - viewbox.min[2]) / viewsize[2] * Menu.sysSize[2]}
		local relMax = {
			(wbox.max[1] - viewbox.min[1]) / viewsize[1] * Menu.sysSize[1],
			(wbox.max[2] - viewbox.min[2]) / viewsize[2] * Menu.sysSize[2]}
		local relSize = {relMax[1]-relMin[1], relMax[2]-relMin[2]}
		Menu:new{
			parent=self;
			isPrompt=true;
			pos=relMin;
			size=relSize;
			entID=entID;
			backgroundColor={0,0,0,0};
			onCancel=menuCloseParent;
			onOk=self.childOnOk;
		}
		if targetID == entID then
			self.client.menuKeyPromptIndex = i
		end
		i=i+1
	end
end
function PickTargetMenu:childOnOk()
	entClientSendMessage(self.client.entID, 'serverBattlePickTarget', 'id', self.entID)
end

--function ClientSide:clientBattlePickTarget(targetting, ...) end

-- this /could/ be done clientside /if/ we gave the client callbacks access to
-- all (visible to their connection) entities rather than just themselves
-- ftw x and y are in world coordinates
function TopviewClientView:serverBattlePickTarget(...)
	local args = {...}
	local battle = assert(self.battle)
	local team = assert(battle:getTeamForClient(self))
	team.pickTargetResult = nil

	assert(#args > 0)
	local picktype = args[1]
	local o
	if picktype == 'pos' then
		local clickpos = vec2:new(
			assert(tonumber(args[2])),
			assert(tonumber(args[3])))
		o = entUnderPos(clickpos)
	elseif picktype == 'id' then
		o = _ents[tonumber(args[2])]
	else
		error()
	end
	
	if o and o:isa(Char) and o.charInfo and o.charInfo.hp > 0 then
		--if o and o is in any hostile team's ent list
		for teamIndex, otherTeam in pairs(battle.teams) do
			--if otherTeam ~= team then	-- targetting dude, targetting
				for _,t in ipairs(otherTeam.ents) do
					if t == o then
						team.pickTargetResult = o
						break
					end
				end
			--end
		end
	end
	
	if team.pickTargetResult then
		self:serverSendMessage('clientBattlePickTargetClose')
	end
end

function ClientSide:clientBattlePickTargetClose()
	-- TODO - pick depending on the client (team) rather than battle
	-- then tell the client to delete its menu
	local m = self.battle.pickTargetMenu
	if m and m.active then m:delete() end
end

	-- BATTLE CMD PICK ITEM

function battleCmdPickItem(cmd, battle, team, char, targetting)
	-- create a menu listing all items, greyd if they can't be used (not battle compat.)
	-- then when its loop is done, return what was chosen
	
	-- how about a generic pick menu
	-- pass a map: item {index, name, number, grey}
	-- then do a forked refresh on a separate thread
	
	team.pickItemResult = nil
	team.client:createPromptingMenu{
		class='PickItemMenu';
	}
	return team.pickItemResult
end;

class('PickItemMenu', Menu)
PickItemMenu.hasPrompts=true
function PickItemMenu:onDelete()
	self.client.battle.pickItemMenu = nil
end
function PickItemMenu:ctor(args)
	args.pos = {1, Menu.sysSize[2] - 11}
	args.size = {Menu.sysSize[1] - 2, 10}
	PickItemMenu.super.ctor(self, args)
	
	self.client.battle.pickItemMenu = self

	-- I'm hijacking this code
	-- maybe I should generalize its naming convention: "refreshItems" instead of having "partyMenu" in there
	fork(function()
		self.client:refreshPartyMenuItems()
		for i,p in ipairs(self.client.partyMenuItems) do
			ItemHolderMenu:new{
				parent=self;
				pos={1, 1.5*i+1};
				itemName=p.name;
				itemIndex=p.index;
				itemAmount=p.amount;
				onOk=self.childOnOk;
				onCancel=menuCloseParent;
			}
		end
	end)
end
function PickItemMenu:childOnOk()
	entClientSendMessage(self.client.entID, 'serverBattlePickItem', self.itemIndex)
end

function TopviewClientView:serverBattlePickItem(itemIndex)
	local battle = assert(self.battle)
	local team = assert(battle:getTeamForClient(self))

	itemIndex = tonumber(itemIndex)
	local item = items[itemIndex]
	if not item then return end
	team.pickItemResult = itemIndex
	self:serverSendMessage('clientBattlePickItemClose')
end

function ClientSide:clientBattlePickItemClose()
	local m = self.battle.pickItemMenu
	if m and m.active then m:delete() end
end

	-- BATTLE CMD PICK SPELL

function battleCmdPickSpell(cmd, battle, team, char, targetting)
	-- create a menu listing all spells, greyd if they can't be used (not enuf mp)
	-- then when its loop is done, return what was chosen
end;

	-- END BATTLE CMD STUFF

function battleItemTargetting(itemIndex)
	return items[itemIndex].targetFlags
end

-- Char:getFightTargetting is related ... the targetting system
function battleSpellTargetting(spellIndex)
	return spells[spellIndex].targetFlags
end

--[[
charInfo ...
where should we go for stats when calc'ing damage?
	... will everyone have a charInfo?
		will random monsters?
		probably not
		so we should either pull from charInfo or from an alternative structure
		or we should always pull from charInfo and draw from charInfo somewhere else for monsters
where should we go for dealing damage?
	... should we put it back in charInfo?
		will that be bad for monsters?
		what about TakesDamage and all the HP stuff already there?
		chances are I'm going to end up circumventing it anyways
--]]
-- [[ todo - put this in "spells.lua" or "attacks.lua" or something
function calcDamage(args)
	local target = assert(args.target)
	local attacker = assert(args.attacker)
	local attackType = assert(args.type)
	local spell = args.spell
	if attackType == 'magic' then assert(spell) end
	local healingAttack = args.healing
	
	local damage
	if attackType == 'magic' then
		local attack = attacker:getStat('magicPower')
		if not attacker.isMonster then
			attack = math.floor(attack * 3/2)
		end
		damage = spell.power * 4 + math.floor(attacker.charInfo.level * attack * spell.power / 32)
		-- if attacker has earrings or hero ring then damage = math.floor(damage * 5 / 4) end
	elseif attackType == 'physical' then
		local attackerBattlePower = attacker:getStat('battlePower')

		local attack
		if attacker.isMonster then
			attack = attackerBattlePower * 4 + math.random(56,64)
		else
			attack = attackerBattlePower + math.min(attacker:getStat('vigor') * 2, 255)
		end
		-- if attacker has gauntlet then attack = attack + math.floor(attackerBattlePower * 3/4) end
		damage = math.floor(attacker.charInfo.level * attacker.charInfo.level * attack / 256)
		if not attacker.isMonster then damage = math.floor(damage * 3/2) + attackerBattlePower end
		-- if attacker has genji glove and attacker is holding 0 or 1 weapons then damage = math.ceil(damage * 3 / 4) end
		-- if attacker has atlas armlet or hero ring then damage = math.floor(damage * 5 / 4) end
	else
		assert(false, "unknown attackType "..attackType)
	end
	-- if #targets > 1 then damage = math.floor(damage / 2) end
	do
		local div = 1
		-- if attacker cmd is 'fight' and attacker is in back row then div = div * 2 end
			-- speaking of which, what's the difference between a 'fight' command any a physical attack?
			-- do non-'fight's (i.e. berzerk) attacks break backrow dividers?
		-- if attacker has hawk eye then div = 1 end
		damage = math.floor(damage / div)
	end
	do
		local mult = 1
		-- if attacker is morphed then mult = mult + 2 end
		-- if attacker is berzerked then mult = mult + 1 end
		-- if attack is a critical hit then mult = mult + 2 end
		damage = damage + math.floor(damage / 2) * mult
	end
	damage = math.floor(damage * math.random(224,255) / 256) + 1
	do
		local defense
		if attackType == 'magic' then
			defense = target:getStat('magicDefense')
		elseif attackType == 'physical' then
			defense = target:getStat('battleDefense')
		else
			assert(false)
		end
		damage = math.floor(damage * (255 - defense) / 256) + 1
	end
	-- if target has safe and attackType == 'physical' then damage = math.floor(damage * 170 / 256) + 1 end
	-- if target has shell and attackType == 'magic' then damage = math.floor(damage * 170 / 256) + 1 end
	-- if target is defending and attackType == 'physical' then damage = math.floor(damage / 2) end
	-- if target is in back row and attackType == 'physical' then damage = math.floor(damage / 2) end
	-- if target is morphed then damage = math.floor(damage / 2) end
	-- if attacker == target and not healingAttack then damage = math.floor(damage / 2) end
	do
		local mult = 0
		-- if we hit target in the back then mult = 1 end
		damage = damage + math.floor(damage / 2) * mult
	end
	-- if target is petrified then damage = 0 end
	-- if target has force field then damage = 0 end
	do
		-- if target absorbs element then healingAttack=true
		-- elseif target is immune to element then damage = 0
		-- elseif target is resistent to element then damage = math.floor(damage/2)
		-- elseif target is weak to element then damage = damage * 2
		-- end
	end
	return damage, healingAttack
end
--]]

function calcFightDamage(target, attacker)
	return calcDamage{
		target=target;
		attacker=attacker;
		type='physical';
	}
end

function calcSpellDamage(target, attacker, spellIndex)
	return calcDamage{
		target=target;
		attacker=attacker;
		type='magic';
		spell=spell[spellIndex];
	}
end

function battleDoFight(target, attacker)
	-- TODO - hit chance
	local damage = calcFightDamage(target, attacker)
	target:takeDamage(attacker, damage)
end

function battleDoItem(target, attacker, itemIndex)
	local item = items[itemIndex]
	item:onUse(target, attacker)
end

function battleDoSpell(target, attacker, spellIndex)
	-- TODO - hit chance, effects, etc
	local damage = calcSpellDamage(target, attacker, spellIndex)
	target:takeDamage(attacker, damage)
end

-- new thread created from the cmd button callback
function runBattleCmd(battle, team, ent, cmd)
	-- then run the command
	local battleCmd = battlecmds[cmd]
	if not battleCmd or not battleCmd.run then
		print('runBattleCmd found an unbound cmd '..tostring(cmd))
	else
		local result = battleCmd:run(battle, team, ent)
		if result then
			-- and then take this person off the ready queue
			team.ready:remove(1)
			-- and reset their battle gauge
			local info = battle.entInfo[ent]
			if info then info.waitGauge = 0 end
		end
	end
	team.client:serverSendMessage('clientBattleCmdDone', cmd)
end

function ClientSide:clientBattleCmdDone(cmd)
	cmd = tonumber(cmd)
	if self.battle and self.battle.cmdMenus then
		local m = self.battle.cmdMenus[cmd]
		if m then m.fontColor = {1,1,1,1} end
	end
end


animLoad('battlecursor', 'battlecursor.png')
entclass('BattleCursor', Renderable)	-- entclass so the classobj is net-registered
BattleCursor.zBias = -10				-- so that, on clientside rendering, the zbias will be picked up
BattleCursor.animGroup = 'battlecursor';



-- battle menu functionality

	--serverside

function TopviewClientView:serverBattlePass()
	local battle = assert(self.battle)
	local team = assert(battle:getTeamForClient(self))
	local ent = team.ready[1]
	team.ready:remove(1)
	team.ready:insert(ent)
end

function TopviewClientView:serverBattleRunCmd(cmd)
	cmd = assert(tonumber(cmd))
	local battle = assert(self.battle)
	local team = assert(battle:getTeamForClient(self))
	-- all of this needs to run in a thread of its own...
	local ent = team.ready[1]
	fork(function()
		runBattleCmd(battle, team, ent, cmd)
	end)
end



function ClientSide:clientBattleClearCurrent()
	if self.battle.currentMenu then
		self.battle.currentMenu:delete()
		self.battle.currentMenu = nil
	end
end

function ClientSide:clientBattleCreateCurrent(...)

	self.battle.currentMenu = Menu:new{
		client=self;
		pos={1,1};
		size={18,8};
		--backgroundColor={0,0,0,.25};
		parent=self.battle.battleMenu;
		hasPrompts=true;
	}
	
	local cmdIndexes = {...}
	for i=1,#cmdIndexes do
		cmdIndexes[i] = tonumber(cmdIndexes[i])
	end
	
	self.battle.cmdMenus = {}
	for i,cmdIndex in ipairs(cmdIndexes) do
		local cmd = battlecmds[cmdIndex]
		if cmd then
			local m = Menu:new{
				client=self;
				pos={0, (i-1)*2};
				size={20, 1.5};
				text=cmd.name;
				parent=self.battle.currentMenu;
				onCancel = function(self, input)
					-- cancel means next entity
					-- i.e. move ready[1] to the end of the list
					-- and close down all associated states (messy with menus atm)
					entClientSendMessage(self.client.entID, 'serverBattlePass')
				end;
				onOk = function(self)
					self.fontColor = {1, .5, 0, 1}
					entClientSendMessage(self.client.entID, 'serverBattleRunCmd', cmdIndex)
				end;
				isPrompt=true;
			}
			self.battle.cmdMenus[cmdIndex] = m
		end
	end
end

-- called from TopviewClientView.render
function ClientSide:updateBattleMenus()
	if self.battle.entMenus then
		for i,entMenu in ipairs(self.battle.entMenus) do
			if i > 4 then break end	-- error?
			local health = net['health'..i]
			local gauge = net['gauge'..i]
			entMenu.health.text = tostring(health)
			entMenu.gauge.size = vec2:new( 10 * gauge / Battle.WAITGAUGEMAX, 1)
			if gauge == Battle.WAITGAUGEMAX then
				entMenu.gauge.backgroundColor = vec4:new(0,1,0,1)
			else
				entMenu.gauge.backgroundColor = vec4:new(1,1,0,1)
			end
		end
	end
end
function ClientSide:clientBattleInit()
	-- out with the old
	self:clientBattleEnd()
	self.updateCallbacks:insertUnique(ClientSide.updateBattleMenus)
	self.battle = {}

	-- in with the new
	self.battle.entMenus = table.new()
	
	self.battle.baseMenu = Menu:new{
		client=self;
		pos={0,0};
		size={Menu.sysSize[1], Menu.sysSize[2]};
		backgroundColor={0,0,0,0};
	}

	self.battle.battleMenu = Menu:new{
		client=self;
		pos={1, Menu.sysSize[2]-11};
		size={Menu.sysSize[1]-2, 10};
		--backgroundColor={0,0,0,.5};
		parent=self.battle.baseMenu;
	}
end

function ClientSide:clientBattleEnd()
	if self.battle then
		if self.battle.baseMenu then self.battle.baseMenu:delete() end
		self.battle = nil
	end
	self.updateCallbacks:removeObject(ClientSide.updateBattleMenus)
end

function ClientSide:clientBattleAddEnt(entId, entName)
	entId = tonumber(entId)

	local i = #self.battle.entMenus+1

	local entMenu = Menu:new{
		client=self;
		pos={22,1+(i-1)*2};
		size={35,2};
		--backgroundColor={0,0,0,.25};
		parent=self.battle.battleMenu;
	}
	local title = Menu:new{
		client=self;
		pos={.25, .25};
		size={13, 1.5};
		--backgroundColor={0,0,0,.75};
		text=entName;
		parent=entMenu;
	}
	local health = Menu:new{
		client=self;
		pos={14,.25};
		size={5,1.5};
		--backgroundColor={0,0,0,.75};
		parent=entMenu;
	}
	local gaugeBack = Menu:new{
		client=self;
		pos={25,.25};
		size={10,1};
		--backgroundColor={0,0,0,0};
		parent=entMenu;
	}
	local gauge = Menu:new{
		client=self;
		pos={0,0};
		size={0,1};
		--backgroundColor={1,1,0,.75};
		borderTex=gaugeBorderTex;
		parent=gaugeBack;
	}
	
	entMenu.health = health
	entMenu.gauge = gauge
	self.battle.entMenus:insert(entMenu)
end



class('Battle')

Battle.WAITGAUGEMAX = 100

function Battle:ctor(args)
	-- shortcut for everything at once
	-- run this on another thread please (just like dialogs and the likes)
	self.teams = table.new()	-- a table of tables

	-- reset everyone's battle gauge by clearing all the entInfo
	self.entInfo = table.new()
	
	-- teams is a key/value mapping from the team key to its members
	for teamIndex, teamarg in pairs(args.teams) do
		--v should be an indexed table of entities
		local team = self:getTeam(teamIndex)
		for _,e in ipairs(teamarg) do
			team.ents:insert(e)
		end
		if teamarg.winCondition then
			team.winCondition = teamarg.winCondition
			-- if it's a single entity and not a list then put it inside a list
			if team.winCondition.isa and team.winCondition:isa(Entity) then
				team.winCondition = table.new{team.winCondition}
			end
		end
		if teamarg.client then
			if teamarg.client.battle then
				print('team '..teamIndex..' client is already in a battle')
				return
			end
			team.client = teamarg.client
			team.client.battle = self
			team.client:serverSendMessage('clientBattleInit', team.client)
			
			local teampos
			local teamdir
			if team.ents and #team.ents > 0 then
				teampos = team.ents[1]:getPos()
				teamdir = team.ents[1].dir
			else
				teampos = team.client:getPos()
			end
			
			-- if we are a client
			-- then we have a party
			-- and one of the party members might be added already? as the char?
			-- which would mean keeping track of who is in and who is out?
			team.client:spawnPartyChars{1,2,3,4,pos=teampos,dir=teamdir}
			
			-- keep track of the indices of 
			team.partyCharsToKeepAfterBattle = table.new()
			
			-- current team
			for _,e in ipairs(team.ents) do
				local charPartyIndex = team.client.partyChars:find(e)
				if charPartyIndex then
					team.partyCharsToKeepAfterBattle[charPartyIndex] = true
				end
				team.client:serverSendMessage('clientBattleAddEnt', e.charInfo.name or '??????')
			end
			
			for charPartyIndex,e in ipairs(team.client.partyChars) do
				if not team.partyCharsToKeepAfterBattle[charPartyIndex] then
					team.ents:insert(e)
					team.client:serverSendMessage('clientBattleAddEnt', e.charInfo.name or '??????')
				end
			end
		end
	end

	self:execute()
	self:shutdown()
end

function Battle:shutdown()
	-- shutdown: kill all menus, cursors, etc
	for _,team in pairs(self.teams) do
		if team.cursor then
			team.cursor:delete()
		end
		if team.client then
		
			local toremove = table.new()
			for charPartyIndex=1,#team.client.partyChars do
				if not team.partyCharsToKeepAfterBattle[charPartyIndex] then
					toremove:insert(charPartyIndex)
				end
			end
			team.client:removePartyChars(toremove)
			
			team.client:serverSendMessage('clientBattleEnd')
			team.client.battle = nil
		end
	end
end

	
function Battle:getTeam(team)	-- creates the team entry if needed.  then returns it
	local t = self.teams[team]
	if not t then
		t = table.new()
		t.ents = table.new()
		t.ready = table.new()
		self.teams[team] = t
	end
	return t
end
	
function Battle:getTeamForClient(client)
	for _,team in pairs(self.teams) do
		if team.client == client then return team end
	end
end
	
function Battle:execute()
	local stillgoing = true
	while stillgoing do
	
		-- process input
		for teamIndex, team in pairs(self.teams) do
			if not team.client then
				-- handle AI
			else
				-- if someone is ready then make sure we have a prompt up for them
				if team.current ~= team.ready[1] then
					-- clear the old menu
					if team.cursor then
						team.cursor:delete()
						team.cursor = nil
					end
					if team.current then
						team.client:serverSendMessage('clientBattleClearCurrent')
					end
					-- set up the current ent
					team.current = team.ready[1]
					if team.current then
						-- put a highlighter above the current ent
						--	TODO - make this only visible to the owning client
						team.cursor = BattleCursor:new()
						team.cursor:setPos( team.current:getPos() + vec2:new(0,-1.2) )
						team.client:serverSendMessage('clientBattleCreateCurrent', unpack(team.current.charInfo.skills))
					end
				end
			end
			
			-- check win condition
			-- or perhaps only do this after a move is made?
			do
				local defeat = table.new()
				if team.winCondition then
					defeat:append(team.winCondition)
				else
					for _,t in pairs(self.teams) do
						if t ~= team then	-- if we're not the active team
							defeat:append(t.ents)
						end
					end
				end

				stillgoing = false
				for _,e in ipairs(defeat) do
					if e.charInfo.hp > 0 then	-- or 'isDead' how about that? then we can abstract health
						stillgoing = true
						break
					end
				end
				if not stillgoing then break end
			end
		end
		
		if not stillgoing then break end
		
		-- advance gauges
		for teamIndex, team in pairs(self.teams) do
			for i,e in ipairs(team.ents) do
			if i > 4 then break end	-- error?
				local info = self.entInfo[e]
				if not info then
					info = {
						waitGauge = 0;
					}
					self.entInfo[e] = info
				end
				if team.client then
					team.client['health'..i] = e.charInfo.hp
				end
				if info.waitGauge < self.WAITGAUGEMAX then
					info.waitGauge = info.waitGauge + (e.battleSpeed or 1)
					if info.waitGauge > self.WAITGAUGEMAX then info.waitGauge = self.WAITGAUGEMAX end
					if team.client then
						team.client['gauge'..i] = info.waitGauge
					end
					if info.waitGauge == self.WAITGAUGEMAX then
						team.ready:insert(e)
					end
				end
			end
		end
		sleep()
	end
end

