require 'wellstory.scripts.ents'
require 'wellstory.scripts.weapons'

-------------------------------- Backpack --------------------------------

--[[
this should be defined before the player is
or at least conveniant so that players can instanciate it per player instance
something like that ...
when a player touches an item, put it in their backpack (and pause the game for a sec)

also in here we want session vars for inter-level-stuff
even though its tempting to write this all off as one big level ...
--]]

class('Backpack')
	-- old system
Backpack.weapon = 0
Backpack.weaponFlags = 0

backpack = Backpack:new()

-------------------------------- Items --------------------------------

entclass('Item', Collidable)

function Item:ctor(so, ...)
	Item.super.ctor(self, so, ...)
		
	-- see if the spawnobj told us to store this in the backpack
	if so.backpackname then self.backpackname = so.backpackname end
	-- classes may override this and explicitly set it in their class

	-- if this is a backpack item
	if self.backpackname then
		-- and if the backpack already has it then dont let it spawn
		if backpack[self.backpackname] then
			-- its not safe to delete in the first frame
			-- or the spawner complains
			-- so instead wait a moment
			setTimeout(function() self:delete() end, 0)
			return
		end
	end
end
	
function Item:touch(other)
	Item.super.touch(self,other)
	if other and other:isa(Player) then
		self:giveItem(other)
		self:delete()
	end
end
	
	-- avoid multiple item class test
function Item:giveItem(other)
	-- keep track of the item if its a backpack one
	if self.backpackname then
		backpack[self.backpackname] = true
	end
end

-------------------------------- Light Scabbard --------------------------------

animLoad('light scabbard', 'lightscabbard.png')

entclass('LightScabbard', Item)
LightScabbard.animGroup = 'light scabbard'
LightScabbard.bbox = box2:new(.9)
LightScabbard.backpackname = 'lightScabbard'
	
function LightScabbard:ctor(so)
	LightScabbard.super.ctor(self,so)
	self:setScale{2,2}
end

function LightScabbard:giveItem(other)
	LightScabbard.super.giveItem(self,other)

	-- the old system
	backpack.weaponFlags = bitor( backpack.weaponFlags, weapons.Sword.weaponFlag )
	backpack.weapon = weapons.Sword.index
end

Spawnable(LightScabbard)

-------------------------------- Super Train --------------------------------

animLoad('heart chest', 'heartchest/heartchest.anim')

-- if you want you can uid these by their spawn object index and map name ...
-- but that'd be a long uid ...
entclass('HeartChest', Item)
HeartChest.animGroup = 'heart chest'
HeartChest.bbox = box2:new(.4)
HeartChest.nextID = 1		--this one is static, so reference as 'HeartChest.nextID' not as a class var
HeartChest.health = 5

function HeartChest:ctor(...)
	-- get our uid
	-- this is a bad idea because, on level chagne, we'll have overlapping uids
	-- solution? explicitly state the id for each item.  tedious.
	self.backpackname = 'heartchest'..HeartChest.nextID
	HeartChest.nextID = HeartChest.nextID + 1

	HeartChest.super.ctor(self, ...)
end
	
function HeartChest:giveItem(other)
	--pause
--		cvarExec('game_time_scalar 0')
	-- run this on another thread - otherwise wait will jank us up
	fork(function()

		-- play some dialogs
		-- sleep is in terms of game time
		-- we need something to wait in terms of system time
		-- egads
		-- i think i'll expose both timers
		other:dialog("YOU'RE HOT")
	
		--hax the plr
		other.maxHealth = other.maxHealth + self.health
		other.health = other.maxHealth

		-- and then give them control back
--			cvarExec('game_time_scalar 1')
	end)
end

Spawnable(HeartChest)
